__title__ = 'mindsdb_sdk'
__package_name__ = 'mindsdb_sdk'
__version__ = '3.1.0'
__description__ = "MindsDB Python SDK, provides an SDK to use a remote mindsdb instance"
__email__ = "jorge@mindsdb.com"
__author__ = 'MindsDB Inc'
__github__ = 'https://github.com/mindsdb/mindsdb_python_sdk'
__pypi__ = 'https://pypi.org/project/mindsdb-sdk/'
__license__ = 'GPL-3.0'
__copyright__ = 'Copyright 2020- mindsdb'
